@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 __version__ = "0.1.19"
 
 import asyncio
@@ -10,16 +12,16 @@ import shutil
 import subprocess
 import sys
 from copy import deepcopy
+from enum import Enum
 from pathlib import Path
-from typing import Any, ClassVar, Dict, List, Optional, Set, Tuple, Union
+from typing import TYPE_CHECKING, Any, ClassVar
 
-if sys.version_info >= (3, 9):
-    from collections.abc import Generator
+if sys.version_info >= (3, 11):
+    from typing import Self
 else:
-    from typing import Generator
+    from typing_extensions import Self
 
 import aiohttp
-import typer
 
 from ._git import clone, fetch
 from ._openfoam import get_changed_binaries, openfoam_version, platform_path
@@ -33,9 +35,12 @@ from ._status import Status
 from ._subprocess import run
 from ._util import path_from_uri, reentrantcontextmanager
 
+if TYPE_CHECKING:
+    from collections.abc import Generator
+
 
 @reentrantcontextmanager
-def _lock() -> Generator[Dict[str, Any], None, None]:
+def _lock() -> Generator[dict[str, Any], None, None]:
     installed_path = platform_path() / "styro" / "installed.json"
 
     installed_path.parent.mkdir(parents=True, exist_ok=True)
@@ -51,12 +56,12 @@ def _lock() -> Generator[Dict[str, Any], None, None]:
         else:
             assert isinstance(installed, dict)
             if installed.get("version") != 1:
-                typer.echo(
+                print(
                     "🛑 Error: installed.json file is of a newer version. Please upgrade styro.",
-                    err=True,
+                    file=sys.stderr,
                 )
                 print_upgrade_instruction()
-                raise typer.Exit(code=1)
+                sys.exit(1)
         installed_copy = deepcopy(installed)
         try:
             yield installed
@@ -80,19 +85,19 @@ class Package:
     )
 
     @staticmethod
-    def _check_for_duplicate_names(pkgs: Set["Package"]) -> None:
+    def _check_for_duplicate_names(pkgs: set[Package]) -> None:
         duplicate_names = {
             pkg.name for pkg in pkgs if len([p for p in pkgs if p.name == pkg.name]) > 1
         }
         if duplicate_names:
-            typer.echo(
+            print(
                 f"🛑 Error: duplicate/conflicting package names: {', '.join(duplicate_names)}",
-                err=True,
+                file=sys.stderr,
             )
-            raise typer.Exit(code=1)
+            sys.exit(1)
 
     @staticmethod
-    def all_installed_binaries() -> Set[Path]:
+    def all_installed_binaries() -> set[Path]:
         with lock as installed:
             return {
                 Path(platform_path() / "bin" / app)
@@ -109,7 +114,7 @@ class Package:
     @staticmethod
     def parse_package(
         package: str,
-    ) -> Union[Tuple[str, None], Tuple[None, str], Tuple[str, str]]:
+    ) -> tuple[str, None] | tuple[None, str] | tuple[str, str]:
         name = package.lower().replace("_", "-")
         if Package.__name_regex.match(name):
             return name, None
@@ -119,28 +124,122 @@ class Package:
             origin = origin.lstrip()
             if Package.__name_regex.match(name):
                 return name, origin
-            typer.echo(
+            print(
                 f"🛑 Error: Invalid package name: {name}",
-                err=True,
+                file=sys.stderr,
             )
-            raise typer.Exit(code=1)
+            sys.exit(1)
         return None, package
 
     @staticmethod
-    def all_installed() -> Set["Package"]:
+    def all_installed() -> set[Package]:
         with lock as installed:
             return {Package(name) for name in installed.get("packages", {})}
 
+    @staticmethod
+    async def _detect_cycles(pkgs: set[Package], *, upgrade: bool = False) -> None:
+        """
+        Detect cycles in the dependency graph before installation.
+
+        Uses depth-first search with three states:
+        - unvisited (white): package not yet visited
+        - visiting (gray): package currently being processed
+        - visited (black): package and all its dependencies processed
+
+        Raises SystemExit if a cycle is detected.
+        """
+
+        class State(Enum):
+            UNVISITED = 0
+            VISITING = 1
+            VISITED = 2
+
+        states: dict[Package, State] = {}
+        path: list[Package] = []
+
+        async def visit(
+            pkg: Package,
+            *,
+            pkg_upgrade: bool = False,
+            pkg_force_reinstall: bool = False,
+        ) -> None:
+            if states.get(pkg, State.UNVISITED) == State.VISITED:
+                return
+
+            if states.get(pkg, State.UNVISITED) == State.VISITING:
+                # Found a cycle - construct the cycle path
+                cycle_start_idx = path.index(pkg)
+                cycle = [*path[cycle_start_idx:], pkg]
+                cycle_names = " -> ".join(p.name for p in cycle)
+
+                print(
+                    f"❌ Dependency cycle detected: {cycle_names}",
+                    file=sys.stderr,
+                )
+                sys.exit(1)
+
+            states[pkg] = State.VISITING
+            path.append(pkg)
+
+            # Follow the same logic as resolve() method
+            # Early return if package is already installed and no upgrade/force reinstall
+            if (
+                pkg.installed_sha() is not None
+                and not pkg_upgrade
+                and not pkg_force_reinstall
+            ):
+                path.pop()
+                states[pkg] = State.VISITED
+                return
+
+            # Check if we need to fetch metadata to get dependencies
+            if pkg._metadata is None:
+                with contextlib.suppress(Exception):
+                    await pkg.fetch()
+
+            # Check again after potential fetch
+            if (
+                pkg._metadata is not None
+                and pkg.installed_sha() is not None
+                and not pkg._upgrade_available
+                and not pkg_force_reinstall
+            ):
+                path.pop()
+                states[pkg] = State.VISITED
+                return
+
+            # Only visit dependencies if the package actually needs resolution
+            # This mirrors the resolve() method logic exactly
+            if pkg._metadata is not None:
+                # Visit requested dependencies (with upgrade=True)
+                for dep in pkg.requested_dependencies():
+                    await visit(dep, pkg_upgrade=True, pkg_force_reinstall=False)
+
+                # Visit installed dependents (reverse dependencies) with force_reinstall=True
+                for dependent in pkg.installed_dependents():
+                    await visit(dependent, pkg_upgrade=False, pkg_force_reinstall=True)
+
+            path.pop()
+            states[pkg] = State.VISITED
+
+        # Start DFS from all root packages with the provided upgrade setting
+        for pkg in pkgs:
+            if states.get(pkg, State.UNVISITED) == State.UNVISITED:
+                await visit(pkg, pkg_upgrade=upgrade, pkg_force_reinstall=False)
+
     @staticmethod
     @lock
     async def resolve_all(
-        pkgs: Set["Package"],
+        pkgs: set[Package],
         *,
         upgrade: bool = False,
-    ) -> Set["Package"]:
+    ) -> set[Package]:
         Package._check_for_duplicate_names(pkgs)
 
-        resolved: Set[Package] = set()
+        # Detect cycles before attempting resolution
+        await Package._detect_cycles(pkgs, upgrade=upgrade)
+
+        resolved: set[Package] = set()
         return {
             pkg
             for pkgs in await asyncio.gather(
@@ -151,7 +250,7 @@ class Package:
 
     @staticmethod
     @lock
-    async def install_all(pkgs: Set["Package"], *, upgrade: bool = False) -> None:
+    async def install_all(pkgs: set[Package], *, upgrade: bool = False) -> None:
         to_install = {
             pkg: asyncio.Event()
             for pkg in await Package.resolve_all(pkgs, upgrade=upgrade)
@@ -171,23 +270,23 @@ class Package:
 
     @staticmethod
     @lock
-    async def uninstall_all(pkgs: Set["Package"]) -> None:
+    async def uninstall_all(pkgs: set[Package]) -> None:
         dependents = set()
         for pkg in pkgs:
             dependents.update(pkg.installed_dependents())
         dependents -= pkgs
         if dependents:
-            typer.echo(
+            print(
                 f"🛑 Error: Cannot uninstall {','.join([pkg.name for pkg in pkgs])}: required by {','.join([dep.name for dep in dependents])}",
-                err=True,
+                file=sys.stderr,
             )
-            raise typer.Exit(code=1)
+            sys.exit(1)
 
         await asyncio.gather(
             *(pkg.uninstall(_force=True) for pkg in pkgs),
         )
 
-    def __new__(cls, package: str) -> "Package":
+    def __new__(cls, package: str) -> Self:
         if cls is not Package:
             return super().__new__(cls)
 
@@ -208,23 +307,23 @@ class Package:
 
     def __init__(self, name: str) -> None:
         if not Package.__name_regex.match(name):
-            typer.echo(
+            print(
                 f"🛑 Error: Invalid package name: {name}",
-                err=True,
+                file=sys.stderr,
             )
-            raise typer.Exit(code=1)
+            sys.exit(1)
         if name == "styro" and not isinstance(self, _Styro):
-            typer.echo(
+            print(
                 "🛑 Error: 'styro' not allowed as a package name.",
-                err=True,
+                file=sys.stderr,
             )
-            raise typer.Exit(code=1)
+            sys.exit(1)
         self.name = name
-        self.origin: Optional[Union[str, Path]] = None
-        self._metadata: Optional[Dict[str, Any]] = None
+        self.origin: str | Path | None = None
+        self._metadata: dict[str, Any] | None = None
         self._upgrade_available = False
 
-    def _build_steps(self) -> List[str]:
+    def _build_steps(self) -> list[str]:
         assert self._metadata is not None
 
         build = self._metadata.get("build", "wmake")
@@ -232,11 +331,11 @@ class Package:
         if build == "wmake":
             build = ["wmake all -j"]
         elif isinstance(build, str):
-            typer.echo(
+            print(
                 f"🛑 Error: Unsupported build system: {build}.",
-                err=True,
+                file=sys.stderr,
             )
-            raise typer.Exit(code=1)
+            sys.exit(1)
 
         return build
 
@@ -266,33 +365,33 @@ class Package:
                     version = int(spec[1:])
                     compatible = openfoam_version() < version
                 else:
-                    typer.echo(
+                    print(
                         f"⚠️ Warning: {self.name}: ignoring invalid version specifier '{spec}'.",
-                        err=True,
+                        file=sys.stderr,
                     )
                     continue
             except ValueError:
-                typer.echo(
+                print(
                     f"⚠️ Warning: {self.name}: ignoring invalid version specifier '{spec}'.",
-                    err=True,
+                    file=sys.stderr,
                 )
                 continue
 
             if (openfoam_version() < 1000) == (version < 1000):  # noqa: PLR2004
                 distro_compatible = True
                 if not compatible:
-                    typer.echo(
+                    print(
                         f"🛑 Error: OpenFOAM version is {openfoam_version()}, but {self.name} requires {spec}.",
-                        err=True,
+                        file=sys.stderr,
                     )
-                    raise typer.Exit(code=1)
+                    sys.exit(1)
 
         if specs and not distro_compatible:
-            typer.echo(
+            print(
                 f"🛑 Error: {self.name} is not compatible with this OpenFOAM distribution (requires {', '.join(specs)}).",
-                err=True,
+                file=sys.stderr,
             )
-            raise typer.Exit(code=1)
+            sys.exit(1)
 
     async def fetch(self) -> None:
         raise NotImplementedError
@@ -302,8 +401,8 @@ class Package:
         *,
         upgrade: bool = False,
         _force_reinstall: bool = False,
-        _resolved: Optional[Set["Package"]] = None,
-    ) -> Set["Package"]:
+        _resolved: set[Package] | None = None,
+    ) -> set[Package]:
         if _resolved is None:
             _resolved = set()
         elif self in _resolved:
@@ -347,7 +446,7 @@ class Package:
     def is_installed(self) -> bool:
         return self in self.all_installed()
 
-    def installed_binaries(self) -> Set[Path]:
+    def installed_binaries(self) -> set[Path]:
         with lock as installed:
             if not self.is_installed():
                 return set()
@@ -364,7 +463,7 @@ class Package:
             except KeyError:
                 return set()
 
-    def installed_sha(self) -> Optional[str]:
+    def installed_sha(self) -> str | None:
         with lock as installed:
             if not self.is_installed():
                 return None
@@ -373,11 +472,11 @@ class Package:
             except KeyError:
                 return None
 
-    def requested_dependencies(self) -> Set["Package"]:
+    def requested_dependencies(self) -> set[Package]:
         assert self._metadata is not None
         return {Package(name) for name in self._metadata.get("requires", [])}
 
-    def installed_dependents(self) -> Set["Package"]:
+    def installed_dependents(self) -> set[Package]:
         with lock as installed:
             return {
                 Package(name)
@@ -389,7 +488,7 @@ class Package:
     def _pkg_path(self) -> Path:
         return platform_path() / "styro" / "pkg" / self.name
 
-    async def download(self) -> Optional[str]:
+    async def download(self) -> str | None:
         raise NotImplementedError
 
     async def install(
@@ -397,7 +496,7 @@ class Package:
         *,
         upgrade: bool = False,
         _force_reinstall: bool = False,
-        _deps: Union[bool, Dict["Package", asyncio.Event]] = True,
+        _deps: bool | dict[Package, asyncio.Event] = True,
     ) -> None:
         with lock as installed:
             if _deps is True:
@@ -410,7 +509,7 @@ class Package:
                 and not upgrade
                 and not _force_reinstall
             ):
-                typer.echo(
+                print(
                     f"✋ Package '{self.name}' is already installed.",
                 )
                 return
@@ -426,7 +525,7 @@ class Package:
                 and not self._upgrade_available
                 and not _force_reinstall
             ):
-                typer.echo(
+                print(
                     f"✋ Package '{self.name}' is already up-to-date.",
                 )
                 return
@@ -466,18 +565,18 @@ class Package:
                                     status=status,
                                 )
                     except subprocess.CalledProcessError as e:
-                        typer.echo(
+                        print(
                             f"🛑 Error: failed to build package '{self.name}'\n{e.stderr}",
-                            err=True,
+                            file=sys.stderr,
                         )
-                        raise typer.Exit(code=1) from e
+                        sys.exit(1)
                     finally:
                         all_installed_binaries = self.all_installed_binaries()
                         for path in list(installed_binaries):
                             if path in all_installed_binaries:
-                                typer.echo(
+                                print(
                                     f"⚠️ Warning: {self.name} modified {path}, which was installed by another package!",
-                                    err=True,
+                                    file=sys.stderr,
                                 )
                                 installed_binaries.remove(path)
 
@@ -525,17 +624,17 @@ class Package:
 
                 self._upgrade_available = False
 
-                typer.echo(f"✅ Package '{self.name}' installed successfully.")
+                print(f"✅ Package '{self.name}' installed successfully.")
 
                 if libs:
-                    typer.echo("⚙️ New libraries:")
+                    print("⚙️ New libraries:")
                     for lib in libs:
-                        typer.echo(f"  {lib}")
+                        print(f"  {lib}")
 
                 if apps:
-                    typer.echo("🖥️ New applications:")
+                    print("🖥️ New applications:")
                     for app in apps:
-                        typer.echo(f"  {app}")
+                        print(f"  {app}")
 
             if isinstance(_deps, dict):
                 _deps[self].set()
@@ -552,9 +651,9 @@ class Package:
 
         with lock as installed:
             if not self.is_installed():
-                typer.echo(
+                print(
                     f"⚠️ Warning: skipping package '{self.name}' as it is not installed.",
-                    err=True,
+                    file=sys.stderr,
                 )
                 return
 
@@ -574,7 +673,7 @@ class Package:
 
         assert not self.is_installed()
 
-        typer.echo(f"🗑️ Package '{self.name}' uninstalled successfully.")
+        print(f"🗑️ Package '{self.name}' uninstalled successfully.")
 
     def __str__(self) -> str:
         return self.name
@@ -592,18 +691,19 @@ class _IndexedPackage(Package):
     async def fetch(self) -> None:
         with Status(f"🔍 Fetching {self}"):
             try:
-                async with aiohttp.ClientSession(
-                    raise_for_status=True
-                ) as session, session.get(
-                    f"https://raw.githubusercontent.com/exasim-project/opi/main/pkg/{self.name}/metadata.json"
-                ) as response:
+                async with (
+                    aiohttp.ClientSession(raise_for_status=True) as session,
+                    session.get(
+                        f"https://raw.githubusercontent.com/exasim-project/opi/main/pkg/{self.name}/metadata.json"
+                    ) as response,
+                ):
                     self._metadata = await response.json(content_type="text/plain")
-            except Exception as e:
-                typer.echo(
+            except Exception as e:  # noqa: BLE001
+                print(
                     f"🛑 Error: Failed to fetch package '{self.name}': {e}",
-                    err=True,
+                    file=sys.stderr,
                 )
-                raise typer.Exit(code=1) from e
+                sys.exit(1)
 
         assert self._metadata is not None
 
@@ -724,8 +824,8 @@ class _Styro(Package):
         *,
         upgrade: bool = False,
         _force_reinstall: bool = False,
-        _resolved: Optional[Set["Package"]] = None,
-    ) -> Set["Package"]:
+        _resolved: set[Package] | None = None,
+    ) -> set[Package]:
         if not upgrade and not _force_reinstall:
             return set()
 
@@ -735,12 +835,12 @@ class _Styro(Package):
             return set()
 
         if is_managed_installation():
-            typer.echo(
+            print(
                 "🛑 Error: this is a managed installation of styro.",
-                err=True,
+                file=sys.stderr,
             )
             print_upgrade_instruction()
-            raise typer.Exit(code=1)
+            sys.exit(1)
 
         return {self}
 
@@ -749,47 +849,47 @@ class _Styro(Package):
         *,
         upgrade: bool = False,
         _force_reinstall: bool = False,
-        _deps: Union[bool, Dict[Package, asyncio.Event]] = True,
+        _deps: bool | dict[Package, asyncio.Event] = True,
     ) -> None:
         if not upgrade and not _force_reinstall:
-            typer.echo(
+            print(
                 "✋ Package 'styro' is already installed.",
             )
             return
 
         if is_managed_installation():
-            typer.echo(
+            print(
                 "🛑 Error: this is a managed installation of styro.",
-                err=True,
+                file=sys.stderr,
             )
             print_upgrade_instruction()
-            raise typer.Exit(code=1)
+            sys.exit(1)
 
         self._upgrade_available = await check_for_new_version(verbose=False)
 
         if not _force_reinstall and not self._upgrade_available:
-            typer.echo(
+            print(
                 "✋ Package 'styro' is already up-to-date.",
             )
             return
 
         await selfupgrade()
 
-        typer.echo("✅ Package 'styro' upgraded successfully.")
+        print("✅ Package 'styro' upgraded successfully.")
 
     async def uninstall(self, *, _force: bool = False, _keep_pkg: bool = False) -> None:
-        typer.echo(
+        print(
             "🛑 Error: styro cannot be uninstalled this way.",
-            err=True,
+            file=sys.stderr,
         )
         if is_managed_installation():
-            typer.echo(
+            print(
                 "💡 Use your package manager (e.g. pip) to uninstall styro.",
-                err=True,
+                file=sys.stderr,
             )
         else:
-            typer.echo(
+            print(
                 "💡 Delete the 'styro' binary in $FOAM_USER_APPBIN to uninstall.",
-                err=True,
+                file=sys.stderr,
             )
-        raise typer.Exit(code=1)
+        sys.exit(1)
